import pickle
import zlib
import ujson as json
import logging

from typing import Any
from time import sleep

class Base(object):

	def _get(self, _:bytes = bytes(), abs:int = 0):
		with self.lock as lock:
			try:
				_ = self.recvall()
			except Exception as e:
				logging.error('Recive faild %s', e)
			else:
				if _ == b'PING':
					self.request.sendall(b'PONG')
				elif _ == b'PONG':
					pass
				elif _ == b'False':
					_ = False
				elif _ == b'True':
					_ = True
				elif _ == b'None':
					_ = None
				elif _ == b'SHUTDOWN':
					if self.connected:
						self.shutdown(2)
				elif _:
					try:
						_ = json.loads(_)
					except Exception as e:
						logging.error('Socket decode faild. %s', e)
						logging.notify(_)
			finally:
				sleep(0)
		# print('got:', _)
		return _

	def _put(self, _: bytes = bytes(), offset: int = 0, **kwargs) -> Any:
		with self.lock as lock:
			if not len(kwargs) and _ == b'':
				_ = None #Can't send a null byte
			if not len(kwargs) and isinstance(_, (bool, type(None))):
				if _ := str(_).encode():
					_ = self.sendall(_)
			elif isinstance(_, bytes) and not len(kwargs):
				_ = self.sendall(_)
			else:
				if _ := json.dumps(_ if _ else kwargs, 
								   separators=(', ', ':')).encode():
					_ = self.sendall(_)
		# print('snd:', _)
		return len(_)



import os
import sys
import logging
import binascii
import bleach

logging = logging.getLogger(__name__)

from flask import (Blueprint,
				   current_app as app, 
				   has_request_context,
				   request,
				   session)

from flask_login import (current_user as worker,
						 user_logged_in,
					     user_logged_out,
					     user_loaded_from_cookie,
					     user_loaded_from_request,
					     user_login_confirmed,
					     user_unauthorized,
					     user_needs_refresh,
					     user_accessed,
					     session_protected)

from ..tcp import TCP
from ..thread import Thread
from .login import AnonymousUserMixin
from .device import Device

bp = os.environ.get('RELOCK_ROUTE', 'relock')
bp = Blueprint(bp, __name__, url_prefix='/%s' % bp,
							 template_folder='templates',
							 static_folder='static',
							 static_url_path='/static/%s' % bp)

class Flask(object):

	def __init__(self, app=None, host=None,
								 port=None,
								 pool=1,
								 ping=False,
								 timeout=30):

		self.host    = str(os.environ.get('RELOCK_SERVICE_HOST', '127.0.0.1'))
		self.port    = int(os.environ.get('RELOCK_SERVICE_PORT', 8111))
		self.pool    = int(os.environ.get('RELOCK_SERVICE_POOL', 1))
		self.ping    = bool(os.environ.get('RELOCK_SERVICE_PING', False))
		self.timeout = int(os.environ.get('RELOCK_SERVICE_TIMEOUT', 30))

		if app is not None:
			self.init_app(app)
		self.tcp = None

	def init_app(self, app, add_context_processor=True):
		""" Configures an application. This registers an `before_request` call, and
			attaches this `Relock service` to it as `app.relock`.

			:param app: The :class:`flask.Flask` object to configure.
			:type app: :class:`flask.Flask`
			:param add_context_processor: Whether to add a context processor to
				the app that adds a `current_user` variable to the template.
				Defaults to ``True``.
			:type add_context_processor: bool
		"""
		app.relock = self

		if not hasattr(app, 'login_manager'):
			raise RuntimeError('Relock service requires Flask-Login to start first.')

		app.login_manager.anonymous_user = AnonymousUserMixin

		app.config.setdefault('RELOCK_SERVICE_HOST', self.host)
		app.config.setdefault('RELOCK_SERVICE_PORT', self.port)
		app.config.setdefault('RELOCK_SERVICE_POOL', self.pool)
		app.config.setdefault('RELOCK_SERVICE_PING', self.ping)
		app.config.setdefault('RELOCK_SERVICE_TIMEOUT', self.timeout)

		with app.app_context():
			try:
				#: Create and initialize the unix sockets TCP client. Client is 
				#: responsible for connection pooling and socket access management.
				self.tcp = TCP(host=app.config.get('RELOCK_SERVICE_HOST'),
							   port=app.config.get('RELOCK_SERVICE_PORT'),
							   pool=app.config.get('RELOCK_SERVICE_POOL'),
							   ping=app.config.get('RELOCK_SERVICE_PING'),
							   timeout=app.config.get('RELOCK_SERVICE_TIMEOUT'))
			except (SystemExit, KeyboardInterrupt):
				sys.exit()
			except Exception as e:
				raise RuntimeError('Relock service host is not available.')
			else:
				#: Setup 'before' and 'after' request calls to the relock service. 
				#: Every new incoming request will create Device object and sends 
				#: the data to relock service.
				app.before_request(Device.before)
				app.after_request(Device.after)
				#: Context methods allows to comunicate with relock service within 
				#: Jinja templates.
				from .context import (x_key_xsid_processor,
									  x_key_screen_processor,
									  x_key_nonce_processor,
									  x_key_signature_processor,
									  x_key_credential_processor,
									  x_key_remote_addr_processor)
				#: Import specific to this Flask plugin routes working inside the 
				#: web application.
				from .routes import (exchange,
									 validate,
									 close,
									 clear,
									 clean,
									 open,
									 js,
									 remote,
									 register,
									 authenticate)
				#: Relock service protects routes from access from 
				#: untrusted/not-registered devices, and as so require to expose
				#: routes that can be accessed public. Expose relock blueprint 
				#: related routes to the relock service.
				app.register_blueprint(bp)
				for p in app.url_map.iter_rules():
					if 'relock' in str(p) and ':' not in str(p):
						self.tcp.expose(str(p))

		#: Expose the main route to the web app.
		self.tcp.expose('/')



import logging
import binascii
from flask import (current_app as app, 
				   has_request_context,
				   request,
				   session)

#: Context functions that can be used directly 
#: by Jinja templates.

@app.context_processor
def x_key_xsid_processor():
	
	def x_key_xsid():
		if hasattr(request, 'device'):
			if xsid := request.device.xsid:
				return xsid
		return str()
	return dict(x_key_xsid=x_key_xsid)

@app.context_processor
def x_key_screen_processor():
	def x_key_screen():
		if hasattr(request, 'device'):
			if screen := request.device.screen:
				return request.device.screen
		return str()
	return dict(x_key_screen=x_key_screen)

@app.context_processor
def x_key_nonce_processor():
	def x_key_nonce():
		if hasattr(request, 'device'):
			return request.device.nonce
		return str()
	return dict(x_key_nonce=x_key_nonce)

@app.context_processor
def x_key_signature_processor():
	def x_key_signature():
		if hasattr(request, 'device'):
			return request.device.signature
		return str()
	return dict(x_key_signature=x_key_signature)

@app.context_processor
def x_key_credential_processor():
	def x_key_credential():
		if hasattr(request, 'device'):
			if credential := request.device.credential:
				return credential
		return str()
	return dict(x_key_credential=x_key_credential)

@app.context_processor
def x_key_remote_addr_processor():
	def x_key_remote_addr():
		if hasattr(request, 'device'):
			if addr := request.remote_addr:
				return addr
		return str()
	return dict(x_key_remote_addr=x_key_remote_addr)


import time
import logging
import threading

from functools import wraps

class Thread(object):

	@classmethod
	def daemon(cls, function):
		@wraps(function)
		def daemon(*args, **kwargs):
			if _ := threading.Thread(target=function, 
									 name=function.__name__, 
									 args=args, 
									 kwargs=kwargs, 
									 daemon=True):
				_.start()
		return daemon

	@classmethod
	def thread(cls, function):
		@wraps(function)
		def thread(*args, **kwargs):
			if _ := threading.Thread(target=function, 
									 name=function.__name__, 
									 args=args,
									 kwargs=kwargs):
				_.start()
				return _.join()
		return thread


import re
import json
import os
import time
import hashlib
import xxhash
import binascii
import base64
import secrets
import logging
import pickle
import zlib

from flask import (current_app as app, 
				   url_for, 
				   redirect,
				   request,
				   abort,
				   has_request_context,
				   session, 
				   Response)
from flask_login import current_user as worker

from datetime import (datetime,
					  timedelta)

from uuid import uuid4
from typing import Any

from ...thread import Thread
from .login import Login

class Device(Login):

	def __init__(self):
		""" The Device object is created with each Flask request and is 
			accessible via a request.device link. Object is responsible 
			for communication with the relock service via TCP Unix sockets.

			The object is initialised before the request processing and 
			it lives as long as the request is being processed.

			Returns:
			    None
		"""
		self.__xsid      = bytes()
		self.__screen    = bytes()
		self.__owner     = False
		self.__nonce     = bytes()
		self.__signature = bytes()
		self.response    = dict()
		self.sid         = str(session.sid)
		self.rid         = str(request.id)
		self.host        = app.config.get('SERVER_HOST')
		#: The relock service TCP socket client registered as a flask 
		#: extension when app was initialized.
		self.relock      = app.relock
		
		if not '/static' in request.path or 'relock.js' in request.path:
			#: Flask request.path has a full url path requested by a user, but 
			#: the relock service requires a version stripped of parameters. 
			#: If possible, all params should be removed to avoid 403 errors 
			#: for forbidden routes.
			if request.url_rule:
				if path := str(request.url_rule) or request.path:
					if request.url_rule:
						if path := re.sub(r'\<.*',"", str(request.url_rule)):
							if len(path) > 1 and path[-1] == '/':
								path = path[:-1]
			else:
				path = request.path			
			#: For each non-static URL path, the relock service requires 
			#: registration of basic information about the request. To simplify 
			#: the management of this data, the initialised object passes the 
			#: required to relock service data by default.
			cookies = dict(session=request.cookies.get('session', str()))
			if request.cookies.get('relock'):
				cookies['relock'] = request.cookies.get('relock')
				cookies['stamp'] = request.cookies.get('stamp')
			with self('before', **{'cookies': cookies,
								   'host': request.headers.get('Host'),
								   'agent': request.headers.get('User-Agent'),
								   'addr': request.remote_addr,
								   'method': request.method,
								   'url': request.url,
								   'path': (path or request.path),
								   'user': worker.get_id(),
								   'email': worker.get_email(),
								   'authenticated': worker.is_authenticated,
								   'active': worker.is_active,
								   'anonymous': worker.is_anonymous,
								   'X-Key-Token': request.headers.get('X-Key-Token', str()) or 
									 		  	  request.form.get('X-Key-Token', str()) or
									 		  	  request.cookies.get('X-Key-Token', str()),
								   'X-Key-Signature': request.headers.get('X-Key-Signature', str()) or 
									 		  	  	  request.form.get('X-Key-Signature', str()) or
									 		  	  	  request.cookies.get('X-Key-Signature', str())}) as tcp:
				if tcp.response in (406, 426):
					#: The server may require to force user to re-authenticate if the 
					#: service-side session expires or the network location changes.
					if worker.is_authenticated:
						worker.logout()
					abort(redirect('/'))
				elif tcp.response in (400, 409, 410):
					#: If server side has key conflict or suspects malicious behaviour 
					#: it may require fresh establish of trust for user. This may require
					#: to cleanup the browser-side keys and new device registration.
					if worker.is_authenticated:
						worker.logout()
					if not path == url_for('relock.clean'):
						abort(redirect(url_for('relock.clean')))
				elif tcp.response in (401, 403):
					#: The response from the service may be forbidden if the 
					#: user keys do not match the server side or if data 
					#: manipulation was detected.
					if worker.is_authenticated:
						worker.logout()
					abort(tcp.response)
				elif tcp.response:
					#: If the request presents valid data that has been verified, 
					#: the server-side response registers in the object key 
					#: parameters. 
					#: 
					#: The key verification is made in separate dedicated 
					#: validation request, as the 'before' route only verificates 
					#: possession of appropriate data, not the ability to rotate 
					#: the transient key. 
					self.__xsid   = tcp.response.get('xsid', str())
					self.__screen = tcp.response.get('screen', str())
					self.__owner  = tcp.response.get('owner', False)
				else:
					#: This can only happen if the TCP socket is broken - it's 
					#: sensible to display 503 error, otherwise the user 
					#: keys may be broken.
					abort(503)

	def __call__(self, route:str = str(), **kwargs):
		""" The object call method is used as a router to the 
			relock service each	time the query to the service is 
			needed.

			:param route: The name of the action on service-side
			:type route: string

			Returns:
			    The service-side generated response for a call.
		"""
		with self.relock.tcp(**{'route': route,
							    'sid': self.sid,
							    'rid': self.rid,
							    'host': self.host,
								**kwargs}) as tcp:
			self.response = tcp.response
		return self

	def __enter__(self):
		return self
 
	def __exit__(self, *args):
		self.response = dict()

	def js(self, id:bytes = bytes(),
				 minified:bool = True,
				 debug:bool = True,
				 host:str = str()):
		""" This method recives the javascript library from the service 
			server. If neccessary the js content may be minified on the 
			relock service side. If debug mode is True the service will 
			check CRC32 sum of the js content and if it change it will 
			send a new version of compiled library otherwise service will 
			provide the checksum of the current version.

			:param id: Checksum of the actuall version of js librarry
			:type id: bytes
			:param minified: Minified version of the javascript requested
			:type minified: boolean
			:param debug: Debug mode allows on-the-fly changes of javascript
			:type debug: boolean
			:param host: Actual service hostname
			:type host: string

			Returns:
			    The server-side generated javascript librarry.
		"""
		with self('js', **{'id': id,
					       'minified': minified,
					       'debug': debug,
						   'host': host}) as tcp:
			if response := tcp.response:
				if response.get('js'):
					response['js'] = bytes(response.get('js', list()))
				return response
		return dict(status=False,
					error='Internal service error.')

	def check(self, token:str = str()) -> bool:
		""" The check method may be used to validate/vertficate the raw 
			token. This method will not validate the signature even if 
			providate as an argument. The token should be providad in a 
			RAW form of bytes or hexlified string. 

			:param token: Token that needs to be validated
			:type token: string

			Returns:
			    Token veryfication boolean confirmation.
		"""
		with self('check', **{'token': token,
							  'reuse': False}) as tcp:
			#: Server side response is always dict with boolean 'status' 
			#: parameter inside. Method itself returns bool True/False 
			#: as a result.
			if tcp.response.get('status'):
				return True
		return dict(status=False,
					error='Internal service error.')

	def confirm(self, reuse:bool = False) -> bool:
		""" This method may be used to validate tokens and signature of 
			token in the same time. The token and signature is not provided 
			directly in arguments but is 'downloaded' from Flask request 
			data. Optional reuse argument switch between alowing the server 
			to reuse token twice, if reuse is False token can be checked 
			only once.

			:param reuse: Allow token reuse, by default the service restricts 
				token verification to one-time use. If reuse is set to True, 
				the relock service will skip checking the previously used list.
			:type reuse: boolean

			Returns:
			    Token veryfication boolean confirmation.
		"""
		with self('confirm', **{'token': request.headers.get('X-Key-Token', str()) or 
								 		 request.form.get('X-Key-Token', str()) or 
								 		 request.cookies.get('X-Key-Token', str()),
							    'signature': request.headers.get('X-Key-Signature', str()) or 
								 		  	 request.form.get('X-Key-Signature', str()) or 
								 		 	 request.cookies.get('X-Key-Signature', str()),
								'reuse': reuse}) as tcp:
			if tcp.response.get('status', False):
				return True
		return False

	def clear(self) -> dict:
		""" This method erase all data about device on the server side. 
			After this command data restore is not possible. 
			
			Server side response does not matter as no data left anyway, 
			so as a	result of this method we got always empty dictionary.
		"""
		with self('clear') as tcp:
			return tcp.response
		return None

	def unlink(self) -> dict:
		""" Same as a clear method, unlink will destroy all informations 
			about the device on server-side. This method is explictly used 
			when user wants to unlink the passkey from the account.
		"""
		with self('unlink') as tcp:
			return None
		return dict(status=False,
					error='Internal service error.')

	def exchange(self, key:bytes = bytes(), 
					   hash:bytes = bytes(), 
					   xsid:bytes = bytes(), 
					   screen:bytes = bytes()) -> dict():
		""" Exchange method is invoked in the moment when a browser is 
			wanted to agree the secret material. The browser generates own 
			private sercret	key material and basis on this public key. 
			Current version of the relock service supports only ECDHE 
			key agreement method.

			If the key agreement is sucessful, the service will return 
			dictionary with a set of the informations:

			key, the public key generated explictly for this browser by a server
			signer, the public key for server signatures 
			token, the set of bytes used by a server to rotate the transient key
			xsid, the hash of the current session id
			recovery, encrypted by a shared key future session key material
			restore, bool information about restoration of the session key
			status, bool information about exchnage process succeed or not
			error, string explanation about eventual failure

			
			:param key: The public key generated by the browser in javascript, 
				the key is dedicated/assigned to this specyfic browser.
			:type key: bytes
			:param hash: The hash of the current version of the transient key, 
				hash is generated by a javascript library.
			:type hash: bytes
			:param xsid: Hash representing current session ID shared by the 
				relock service and browser. Hash is constant to the server side
				session and provided by a relock service to the browser.
			:type xsid: bytes
			:param screen: Hash associated with the current tab opened by a 
				user's browser. This hash is automatically generated by a 
				javascript when new browser session storage is sandboxed.
			:type screen: bytes
		"""
		with self('exchange', **{'key': key,
							     'hash': hash,
							     'xsid': xsid,
							     'screen': screen}) as tcp:
			return tcp.response
		return dict(status=False,
					error='Internal service error.')

	def validate(self, screen:bytes = bytes(), 
					   nonce:bytes = bytes(), 
					   token:[str|bytes] = str(), 
					   signature:[str|bytes] = str()) -> dict:
		""" The validate method is invoked by a browser every time when the key
			has been rotated and/or needs confirmation. As a result the method 
			returns a dictionary with a set of informations:

			status, boolean value representing the state of the process
			authenticated, bool information about user authentication status
			credential, bool confirmation that passkey is assigned to the device
			url, string url address required to be redirected after the process
			reprocess, bool indicator is fresh registration possible or not
			timeout, int value about required delay before redirect
			owner, bool confimration that device has been assigned to the owner

			:param screen: Dedicated hash generated by the browser for the 
				currently in use browser tab.
			:type screen: bytes
			:param nonce: Random set of bytes used to renew the secret material 
				of the transient key
			:type nonce: bytes
			:param token: The set of bytes generated basis on the current 
				version of the transient key. Defaults hexlified to the form of 
				the str object.
			:type token: bytes
			:param signature: The set of bytes generated using private key and 
				token that confirms the token comes from legitimate source.
				By default hexlified to the form of the str object.
			:type signature: bytes
		"""
		with self('validate', **{'screen': screen,
							     'nonce': nonce,
							     'token': token,
							     'signature': signature}) as tcp:
			if response := tcp.response:
				if not response.get('status'):
					#: The False status means we can't confirm the session and
					#: confirm device keys. To ensure secure state, user session
					#: should be revoked imidietly.
					if worker.is_authenticated:
						worker.logout()
				return response
		return dict(status=False)

	def webauthn(self, options:dict = dict()) -> dict:
		""" Generate options for registering a credential via 
			navigator.credentials.create()

			Returns:
			    Registration options ready for the browser. Consider 
			    using `helpers.options_to_json()` in this library to 
			    quickly convert the options to JSON.
		"""
		with self('webauthn', **{'options': options}) as tcp:
			return tcp.response

	def authenticate(self, credential:dict = dict()) -> dict:
		""" Generate options for authentication and also if credential 
			resolving the challange by a browser is passed authenticates 
			the user to the system.

			Returns:
				Dictionary for a stringified JSON version of the 
				WebAuthn response.
		"""
		with self('credential', **{'credential': credential}) as tcp:
			return tcp.response

	@property
	def credential(self) -> bool:
		""" Check if actual device has registered a corresponding
			credential for passkey authentication. If credentail is 
			assigned returns True.
		"""
		with self('credential') as tcp:
			return bool(tcp.response)
		return False

	@Thread.daemon
	def open(self, screen:str = str(), 
				   origin:str = str(), 
				   path:str = str(),
				   xsid:str = str(),
				   server:str = str()) -> dict:
		""" ...

			:param screen: Hash of the actuall tab id openend in browser
			:type screen: string
			:param origin: Hostname passed by a javascript librarry
			:type origin: string
			:param path: Document location passed by a js lib
			:type path: string

			Returns:
				None
		"""
		with self('open', **{'screen': screen,
							 'origin': origin,
							 'path': path,
							 'xsid': xsid,
							 'server': server}) as tcp:
			return None

	@Thread.daemon
	def close(self, screen:str = str(), 
					origin:str = str(), 
					path:str = str()) -> dict:
		""" Each time the browser unloads the web page content from a 
			sandbox, the JavaScript library sends a beacon with information 
			about unloaded page. This method can be called asynchronously 
			as it does not return a response.

			:param screen: Hash of the actuall tab id openend in browser
			:type screen: string
			:param origin: Hostname passed by a javascript librarry
			:type origin: string
			:param path: Document location passed by a js lib
			:type path: string

			Returns:
				None
		"""
		with self('close', **{'screen': screen,
							  'origin': origin,
							  'path': path}) as tcp:
			return None

	def remote(self, screen:str = str()) -> dict:
		""" Demo purpose only. This method emulates compromised key 
			usage by a remote server.
		"""
		with self('remote', **{'screen': screen}) as tcp:
			return tcp.response

	def sign(self, value:bytes) -> bool:
		""" This method signs the data using dedicated for the device
			private key.
		"""
		with self('sign', **{'value': value}) as tcp:
			if tcp.response:
				return tcp.response
		return None

	def token(self) -> bytes:
		""" This method generates a fresh token using current version of 
			the server-side transient key. Multiple tokens per request 
			may be generated. 
		"""
		with self('token') as tcp:
			if tcp.response:
				return tcp.response
		return None

	def verify(self, value:bytes = bytes(), 
					 signature:bytes = bytes()) -> bool:
		""" This method validates the token and a signature passed as 
			parameters. If the signature is appropriate and the token 
			matches the server-side transient key, it returns True. 

			:param value: Token that has to be verificated
			:type value: bytes
			:param signature: Token signature by a corresponding private key
			:type signature: bytes
		"""
		with self('verify', **{'value': value,
							   'signature': signature}) as tcp:
			if tcp.response:
				return tcp.response
		return None

	@property
	def protected(self):
		""" Setup a protected mode for the device. If protected mode is
			set to True, relock service will not allow to create any
			new device for user account and all request to the application
			will be allowed only from authorized devices only.

			This turns on the strict device veryfication mode.
		"""
		with self('protected', **{'state': None,
								  'user': session.get('_user_id'),
								  'email': session.get('email')}) as tcp:
			return bool(tcp.response)
		return False

	@protected.setter
	def protected(self, state:bool = None):
		with self('protected', **{'state': bool(state),
								  'user': session.get('_user_id'),
								  'email': session.get('email')}) as tcp:
			if tcp.response:
				return tcp.response
		return None

	@property
	def owner(self):
		""" The x-owner is generated along with the object initialization, 
			if attribute exists and server-side Device object is assigned 
			to this practicular id, the property returns exactly the same 
			id as passed to init of the object.

			Returns:
				The user identity registered to the device.
		"""
		return self.__owner

	@property
	def screen(self):
		""" The x-screen is generated along with the object initialization, 
			if attribute exists, the property returns a hexadecimal version 
			of the bytes. It represents the actual, random, identity of the 
			user tab in the browser.

			Returns:
				Hexlified bytes representing server-side screen id.
		"""
		return self.__screen

	@property
	def xsid(self):
		""" The x-session-id is generated along with the object initialization, 
			if attribute exists, the property returns a hexadecimal version of 
			the bytes.

			Returns:
				Hexlified bytes representing server-side session id.
		"""
		return self.__xsid

	@property
	def signature(self):
		""" The signature is generated along with the nonce, if exists, 
			the property returns a hexadecimal version of the signature 
			bytes.

			Returns:
				Hexlified bytes representing nonce signature.
		"""
		if self.__nonce and self.__signature:
			return binascii.hexlify(self.__signature).decode()
		return str()

	@property
	def nonce(self) -> str:
		""" The nonce property is used manually to trigger the re-keying of 
			the transient key on the selected page views. Once triggered the
			nonce will be used in key rotation before validation request.

			Returns:
				Hexlified bytes generated for re-keying / nonce bytes.
		"""
		if not self.__nonce:
			#: re-keying nonce may be generated only one-time per request
			#: processing, so there is no sense to ask relock service
			#: multiple times about it.
			with self('nonce') as tcp:
				if tcp.response:
					print(tcp.response)
					if nonce := binascii.unhexlify(tcp.response.get('nonce')):
						self.__nonce = nonce
						if signature := binascii.unhexlify(tcp.response.get('signature')):
							self.__signature = signature
		if self.__nonce:
			return binascii.hexlify(self.__nonce).decode()
		return str()

	""" Before processing the request, we need to initialise the Device object 
		on the application side. The application will have access to the relock 
		service via the TCP socket handled by this connector.

		The relock service require two identifaction items, the session Id 
		and the rquest Id to process the request correctly.

		Returns:
			None
	"""
	@classmethod
	def before(cls) -> None:
		#: If current session has nothing in a server-side storage, Flask 
		#: will create a new fresh session id for each new incoming request. 
		#: Thats the reason of creating the random ['id'] in session storage.
		if not 'id' in session:
			session['id'] = str(uuid4())
		#: If request has no any identyficator that we can use it, let's 
		#: create one before sending data to relock service.
		if not hasattr(request, 'id'):
			setattr(request, 'id', str(uuid4()))
		#: Durring the request processing we can use the same/single TCP
		#: pipeline, that will speed up relock-side processing.
		if not hasattr(request, 'device'):
			if request.method.lower() in ('get', 'post'):
				setattr(request, 'device', Device())


	"""	After processing of the request the relock service may return cookie
		directives (set or delete cookie).

		:param response: Response generated by the Flask framework for this 
			practicular request.
		:type response: Flask response object

		Returns:
			Flask response object
	"""
	@classmethod
	def after(cls, response) -> object:
		#: Request finalization may return information about cookies needed 
		#: to be set after request, however there is no need to process
		#: anyting if we are processing the /static path. Static files don't
		#: set cookies.
		if request.method.lower() in ('get', 'post'):
			if not '/static' in request.path:
				#: Flask is closing TCP connection after request processing so
				#: it's required to use fresh pipeline to finalize the operation.
				#: New pipe may not neccessary be connected to the same relock 
				#: service, and request may be processed by different server.
				with app.relock.tcp(**{'route': 'after',
									   'sid': session.sid,
									   'rid': request.id,
									   'status': response.status,
									   'code': response.status_code,
								  	   'host': app.config.get('SERVER_HOST')}) as tcp:
					if tcp.response:
						#: The cookie data coming from relock service are raw, so 
						#: we need to convert seconds to the apropriate timestamp 
						#: for browser cookie set.
						for name, cookie in tcp.response.items():
							if 'value' in cookie:
								response.set_cookie(name, value=cookie.get('value'),
														  expires=datetime.now() + timedelta(seconds=cookie.get('expires')),
														  max_age=timedelta(seconds=cookie.get('max_age')),
														  path=cookie.get('path'),
														  domain=cookie.get('domain'),
														  secure=cookie.get('secure'), 
														  httponly=cookie.get('httponly'),
														  samesite=cookie.get('samesite'))
							else:
								#: If cookie has no any value this means relock service
								#: require cookie delete.
								response.delete_cookie(name, path='/')
		return response




import binascii
import logging
import time

from flask import (current_app as app,
				   request,
				   Response,
				   session,
				   redirect,
				   url_for,
				   abort)

from flask_login import (current_user as worker,
						 logout_user,
						 user_logged_in,
						 user_logged_out,
						 user_loaded_from_cookie,
						 user_loaded_from_request,
						 user_login_confirmed,
						 user_unauthorized,
						 user_needs_refresh,
						 user_accessed,
						 session_protected)

from ...thread import Thread

class Login(object):

	#: Sent when a user is logged in. In addition to the app (which is the
	#: sender), it is passed `user`, which is the user being logged in.
	@user_logged_in.connect
	def _user_logged_in(self, user):
		with self.relock.tcp(**{'route': 'user_logged_in',
							    'sid': session.sid,
							    'rid': request.id,
								'user': user.get_id(),
						        'email': user.email,
						        'authenticated': user.is_authenticated,
						        'active': user.is_active,
						        'anonymous': user.is_anonymous,						    
							    'host': app.config.get('SERVER_HOST')}) as tcp:
			if tcp.response in (423, 410):
				if '_user_id' in session:
					del session['_user_id']

	#: Sent when a user is logged out. In addition to the app (which is the
	#: sender), it is passed `user`, which is the user being logged out.
	@user_logged_out.connect
	def _user_logged_out(self, *args, **kwargs):
		with self.relock.tcp(**{'route': 'user_logged_out',
							    'sid': session.sid,
							    'rid': request.id,
							    'addr': request.remote_addr,
							    'host': app.config.get('SERVER_HOST')}) as tcp:
			if tcp.response in (423, 410):
				if '_user_id' in session:
					del session['_user_id']

	#: Sent when the user is loaded from the cookie. In addition to the app (which
	#: is the sender), it is passed `user`, which is the user being reloaded.
	@user_loaded_from_cookie.connect
	def _user_loaded_from_cookie(self, *args, **kwargs):
		logging.debug('_user_loaded_from_cookie')

	#: Sent when the user is loaded from the request. In addition to the app (which
	#: is the #: sender), it is passed `user`, which is the user being reloaded.
	@user_loaded_from_request.connect
	def _user_loaded_from_request(self, *args, **kwargs):
		logging.debug('user_loaded_from_request')

	#: Sent when a user's login is confirmed, marking it as fresh. (It is not
	#: called for a normal login.)
	#: It receives no additional arguments besides the app.
	@user_login_confirmed.connect
	def _user_login_confirmed(self, *args, **kwargs):
		logging.debug('_user_login_confirmed')

	#: Sent when the `unauthorized` method is called on a `LoginManager`. It
	#: receives no additional arguments besides the app.
	@user_unauthorized.connect
	def _user_unauthorized(self, *args, **kwargs):
		logging.debug('user_is_unauthorized_to_get_response')
		
	#: Sent when the `needs_refresh` method is called on a `LoginManager`. It
	#: receives no additional arguments besides the app.
	@user_needs_refresh.connect
	def _user_needs_refresh(self, *args, **kwargs):
		logging.debug('_user_needs_refresh')

	#: Sent whenever the user is accessed/loaded
	#: receives no additional arguments besides the app.
	@user_accessed.connect
	def _user_accessed(self, *args, **kwargs):
		session.modified = True
		#: send the beacon only when user is authenticated
		#: overwise is not necessary to waste computing power
		if session.get('_user_id'):
			if not '/static' in request.url:
				with self.relock.tcp(**{'route': 'user_accessed',
									    'sid': session.sid,
									    'rid': request.id,
									    'host': app.config.get('SERVER_HOST')}) as tcp:
					if tcp.response in (423, 410):
						del session['_user_id']

	#: Sent whenever session protection takes effect, and a session is either
	#: marked non-fresh or deleted. It receives no additional arguments besides
	#: the app.
	@session_protected.connect
	def _session_protected(self, *args, **kwargs):
		pass



__author__ = 'Marcin Sznyra'
__credits__ = 'relock Inc.'

from .tcp import TCP
from .tcp.socket import Socket
from .thread import Thread
try:
	from flask import session
except Exception as e:
	pass
else:
	from .flask import Flask


import time
import logging
import binascii
import ujson as json

from typing import Any

from ..thread import Thread
from .pool import Pool

class Events(object):

	def __init__(self):
		pass

	def timeadd(self, key:str, **kwargs):
		"""Event creates of a new thread-specific connection, and send request to put in a sorted set a new item with value passed in kwargs argument collection. Sorted set is created by using current timestamp.

		Args:
		    `key`: A unique, constant identifier for database identyfication.
		    (optional) `kwargs`: A collection of any kind key/value pairs saved together with unqiue `key`.
		Returns:
		    None
		"""
		if conn := abs(self):
			if conn._put(route = 'zadd', 
						 key = key,
						 score = time.time(),
						 value = kwargs):
				return conn._get()

	def notify(self, **kwargs):
		"""
		Event creates of a new thread-specific connection, and send request to 
		sentinel with a value passed in kwargs argument collection.

		Args:
		    `kwargs`: A collection of any kind key/value pairs.
		Returns:
		    None
		"""
		with self('notify', **kwargs) as self:
			return self.response

	def expose(self, url):
		"""
		Send to the relock exposed route address/name and register as
		unprotected route.

		Args:
		    `kwargs`: A collection of any kind key/value pairs.
		Returns:
		    None
		"""
		with self('expose', **{'url': url}) as self:
			return self.response




import time

class AnonymousUserMixin:
	"""
	This is the default object for representing an anonymous user.
	"""
	def __enter__(self):
		return self
 
	def __exit__(self, *args):
		pass

	@property
	def is_authenticated(self):
		return False

	@property
	def is_active(self):
		return False

	@property
	def is_anonymous(self):
		return True

	@property
	def veryficated(self):
		return False

	def get_id(self):
		return None

	def get_email(self):
		return str()

	def check_password(self, _):
		return False

	def remove(self):
		return False

	def __abs__(self):
		return 0

	def __int__(self):
		return 0

	def __str__(self):
		return str(self.get_email())


import time
import logging
import sys
import socket
import signal

from dataclasses import dataclass
from contextlib import closing

from .socket import Socket

@dataclass
class Server:
	
	host: str        = str()
	port: int 	     = 0
	ping: bool       = False
	pool: object     = None

	def __abs__(self):
		with self.pool as conn:
			return bool(conn)
		# with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
		# 	if sock.connect_ex((self.host, self.port)) == 0:
		# 		return True
		return False

class Pool(list):

	length: int  	 = 2048
	ping: bool  	 = False
	size: int 		 = 1
	host: str   	 = str()
	port: int   	 = 0
	expire: int 	 = 60

	__id__: int 	 = 0
	__cn__: int 	 = 1

	def __init__(self, host:str    = str(), 
					   port:int    = int(), 
					   pool:int    = 1, 
					   ping:bool   = False,
					   lock:object = None,
					   expire:int  = 60):
		Pool.host   = host
		Pool.port   = port
		Pool.pool   = pool
		Pool.ping   = ping
		Pool.lock   = lock
		Pool.expire = expire
		for x in range(pool):
			self(host, port)

	def __enter__(self):
		return next(self)

	def __exit__(self, *args):
		pass

	def __call__(self, host:str, port:int):
		if id := self.__cn__ + 1:
			self.__cn__ = id

			if conn := Socket(host, port, 
									expire=self.expire,
									lock=self.lock,
								    id=id):
				self.append(conn)
			# with self[-1] as conn:
		return self[-1]

	def __iter__(self):
		for i in range(len(self)):
			yield self[i]

	def __next__(self):
		if self.__id__ == len(self):
			self.__id__ = 0
		self.__id__ = self.__id__ + 1
		if conn := self[self.__id__ - 1]:
			# if time.time() > conn.expire:
			# 	return next(self.restore(conn))
			return conn
		if len(self):
			if conn := self.restore(self[self.__id__ - 1]):
				return next(self)
		raise ConnectionRefusedError('Connection has been lost. %s:%s:%s', self.__id__, len(self), self[0])

	def __bool__(self):
		return True if len(self) else False

	def shutdown(self, conn):
		if conn in self:
			super().remove(conn)
		self.__cn__ = self.__cn__ - 1
		del conn
		return self

	def restore(self, conn):
		if self(*conn.addr):
			self.shutdown(conn)
		return self

class Cluster(list):

	ping: bool  = False
	pool: int 	= 1

	__id__: int = 0

	def __init__(self, pool:int = 1, 
					   ping:bool = False,
					   lock:object = None):
		self.pool = int(pool)
		self.ping = bool(ping)
		self.lock = lock

	def __enter__(self):
		return next(self)

	def __exit__(self, *args):
		pass

	def __call__(self, host:str, port:int):
		if not (host, port) in self:
			if _ := Server(host,
						   port,
						   self.ping,
						   Pool(host, 
						   		port, 
						   		self.pool, 
						   		self.ping,
						   		self.lock)):
				if abs(_):
					self.append(_)
					return _

	def __iter__(self):
		for i in range(len(self)):
			yield self[i]

	def __contains__(self, addr:tuple):
		if isinstance(addr, tuple):
			for server in self:
				if server.host == addr[0] and server.port == addr[1]:
					return True
		else:
			for server in self:
				if server == addr:
					return True
		return False

	def __next__(self):
		if self.__id__ >= len(self):
			self.__id__ = 0
		self.__id__ += 1
		if _ := len(self):
			if server := self[self.__id__ - 1]:
				# if abs(server):
				return server

	def __bool__(self):
		# for server in self:
		# 	if not abs(server):
		# 		self.servers.remove(server)
		return True if len(self) else False

	def remove(self, item):
		# if not len(self):
		# 	raise ValueError('No active servers left.')
		for server in list(self):
			if server == item:
				super().remove(server)	
			elif server.host == item.host and server.port == item.port:
				super().remove(server)


import logging

from typing import Any

from .events import Events

class Core(Events):

	def __init__(self):
		pass

	def get(self, key:str):
		with self('get', key=key) as _:
			return self.response

	def set(self, key:str, value:Any):
		with self('set', key=key, value=value) as self:
			return self.response

	def delete(self, key:str):
		with self('delete', key=key) as self:
			return self.response

	def exists(self, key:str):
		with self('exists', key=key) as self:
			return self.response

	def keys(self, key:str):
		with self('keys', key=key) as self:
			return self.response

	def ttl(self, key:str, value:int):
		with self('ttl', key=key, value=value) as self:
			return self.response

	def expire(self, key:str, value:int):
		with self('expire', key=key, value=value) as self:
			return self.response

	def zadd(self, key:str, score:int, **kwargs):
		""" Create new connection for thread
		"""
		with self('zadd', key=key,
						  score=score,
						  value=kwargs) as self:
			return self.response

	def zrange(self, key:str, x:int=0, y:int=-1):
		with self('zrange', key=key, x=0, y=-1) as self:
			return self.response

	def zrevrange(self, key:str, x:int=0, y:int=-1):
		with self('zrevrange', key=key, x=0, y=-1) as self:
			return self.response

	def zrem(self, key:str, value:Any):
		with self('zrem', key=key, value=value) as self:
			return self.response

	def sadd(self, key:str, **kwargs):
		with self('sadd', key=key, **kwargs) as self:
			return self.response

	def srem(self, key:str, **kwargs):
		with self('srem', key=key, **kwargs) as self:
			return self.response

	def smembers(self, key:str, **kwargs):
		with self('smembers', key=key) as self:
			return self.response



import time
import logging
import sys
import socket
import signal

from ctypes import c_ulong
from fcntl import ioctl

from .base import Base

from gevent import sleep

class Socket(Base):

	length:int  	 = 2048
	connected:bool 	 = False
	_bytes:int 		 = 4

	def __init__(self, host:str, 
					   port:str, 
					   lock:object = None,
					   **kwargs):

		self.__id__		  = kwargs.get('id', 0)
		self.__response	  = None
		self.addr         = (host, port)
		self.lock         = lock
		self.expire		  = time.time() + kwargs.get('expire', 3600)
		self.connected    = False

		self.request = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.request.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.request.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
		# self.request.setblocking(0)

	def __call__(self, route:str, **kwargs):
		if self._put(**{'route': route, **kwargs}):
			self.__response = self._get()
		return self

	def __enter__(self):
		if not bool(self):
			return abs(self)
		return self

	def __exit__(self, *args):
		self.__response = None

	def __abs__(self, _:bytes = bytes()):
		try:
			self.request.connect(self.addr)
		except ConnectionRefusedError:
			logging.notify('Connection Refused %s:%s, host is down.', *self.addr)
			raise ConnectionRefusedError('Host %s:%s is down.' % self.addr)
		except TimeoutError:
			logging.notify('Timeout error, host %s:%s is down.', *self.addr)
			raise ConnectionRefusedError('Host %s:%s is down.' % self.addr)
		else:
			self.connected = True
		return self

	def __bool__(self, _:bool = False, bytes:bytes = bytes()) -> bool:
		if not self.connected:
			abs(self)
		try:
			if self.sendall(b'PING'):
				bytes = self.recvall(socket.MSG_DONTWAIT)
		except ConnectionRefusedError: #host is down
			logging.error('ConnectionRefusedError')
			_ = False
		except BlockingIOError:
			logging.error('BlockingIOError')
			_ = True
		except OSError as e: #not connected yet
			logging.error(e)
			_ = False
		except ConnectionResetError: #broken pipe
			logging.error('ConnectionResetError')
			_ = False
		except socket.timeout as e:
			logging.error('SocketTimouetError')
			_ = False
		except Exception as e:
			logging.exception("Unexpected problem when checking socket connection.")
			_ = False
		else:
			if bytes == b'PONG':
				_ = True
		finally:
			self.connected = _
		return self.connected

	@property
	def id(self):
		return self.__id__

	@id.setter
	def id(self, value:int):
		self.__id__ = value

	@property
	def response(self):
		return self.__response

	def recv(self):
		return self.request.recv(self.length)

	def send(self, value:bytes):
		if not self.request._closed:
			return self.request.send(value)

	def sendall(self, _:bytes = bytes()):
		if abs := len(_).to_bytes(self._bytes, byteorder='big'):
			self.request.sendall(abs + _)
		return _

	def recvall(self, *flags, _:bytes = bytes()):
		if abs := self.request.recv(self._bytes, *flags):
			if abs := int.from_bytes(abs, byteorder="big"):
				while slice := self.request.recv(self.length, *flags):
					_ += slice
					if len(_) >= abs:
						break
		return _

	def shutdown(self, how=0):
		logging.notify(
				"Client closing connection from %s:%s",
				*self.addr)
		try:
			if not self.request._closed:
				self.request.shutdown(how)
		except OSError:
			self.close()

	def close(self):
		try:
			if not self.request._closed:
				self.request.close()
		except Exception as e:
			logging.notify(e)
		finally:
			self.disconnected()

	def disconnected(self):
		if hasattr(self, 'addr'):
			logging.notify('Client disconnected from server %s:%s', *self.addr)

	def __delete__(self):
		try:
			if not self.request._closed:
				self.request.shutdown(0)
		except Exception as e:
			logging.warning(e)
		else:
			self.request.close()
		finally:
			logging.info('Closed connection to server %s%s:', self.addr)

	def flush(self) -> bool:
		return False


import sys
import time, logging
import requests
import pickle
import binascii
import hashlib
import base64
import socket

logging = logging.getLogger('sentinel.tcp.client')

from typing import Any
from uuid import uuid4

from time import sleep

from ..thread import Thread

from .base import Base
from .pool import Pool, Cluster
from .core import Core

from threading import Lock

class TCP(Core, Base):

	_response: Any = None

	def __init__(self, host: list  = list(), 
					   port: int   = 8111,
					   pool: int   = 1,
					   ping: bool  = False,
					   timeout:int = 60,
					   schema:str  = 'tcp'):
		self.id = str(uuid4())
		self.servers = Cluster(pool, 
							   ping, 
							   Lock())
		if not isinstance(host, list):
			host = [(host, int(port)),]
		for host, port in host:
			self.servers(host, port)
		if round(self):
			super().__init__()
		self.refresh_sentinel_tenants(timeout)

	def __call__(self, route:str, **kwargs):
		if self.servers:
			with self.servers as server:
				try:
					with server.pool as conn:
						if conn and conn._put(**{'route': route, **kwargs}):
							self._response = conn._get()
						else:
							logging.notify('Sentinel connection has gone. Rounding.')
							if round(self, server):
								return self(route, **kwargs)
				except (IndexError, OSError, ConnectionRefusedError):
					logging.debug('Route to Sentinel no longer exists, host {blue}%s:%s{z}{g} have gone down.', server.host, server.port)
					if round(self, server):
						return self(route, **kwargs)
				except Exception as e:
					logging.error('Unidentified fault %s', e)
		return self

	def __abs__(self):
		with self.servers as server:
			if _ := server.pool(server.host, server.port):
				return _

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self._response = None
		sleep(0)

	def __iter__(self):
		for i in range(len(self.servers)):
			yield self.servers[i]

	def __round__(self, server:str = None):
		if server is not None:
			self.servers.remove(server)
			logging.notify('Serever %s:%s is down', server.host, server.port)
		if self.servers:
			with self('members') as self:
				for id, sentinel in self.response.items():
					if not (sentinel.get('addr'),
							sentinel.get('port')) in self.servers:
						try:
							if server := self.servers(sentinel.get('addr'), 
										 			  sentinel.get('port')):
								logging.debug('New server %s:%s in pool', server.host, server.port)
						except:
							#: Remove an unreachable server from the ring
							with self.servers as server:
								try:
									with server.pool as conn:
										if conn and conn._put(**{'route': 'missing', **sentinel}):
											_response = conn._get()
								except:
									pass
				logging.debug('Rounding routes, available: %s', len(self.servers))
		return len(self.servers)

	@property
	def response(self):
		return self._response

	def shutdown(self, how):
		logging.info('Shutdown requested %s', how)
		self._response = None
		self.request.shutdown(how)
		self.request.close()

	@Thread.daemon
	def refresh_sentinel_tenants(self, timeout):
		sleep(timeout);
		if round(self):
			return self.refresh_sentinel_tenants(timeout)



from flask import (current_app as app, 
				   session, 
				   request, 
				   Response, 
				   url_for, 
				   json,
				   render_template, 
				   flash, 
				   jsonify, 
				   redirect, 
				   abort, 
				   make_response, 
				   session)

from flask_login import (current_user as worker, 
						 login_required)
from . import bp, logging

import bleach
import random, os
import time
import zstd
import pickle
import base64
import hashlib
import binascii
import subprocess

from datetime import datetime
from datetime import timedelta
from jsmin import jsmin
from gevent import sleep

from urllib.parse import urlparse

@bp.route('/remote', methods=['POST'])
def remote():
	""" This is a demo purpose only method. It should not be implemented 
		in production ready systems. The method when invoked (and should be
		invoked manually) emulates the behavior of stolen keys attack.

		This route is invoked by a user manually by a pressing the button
		inside the demo application.
	"""
	if response := request.device.remote():
		return response
	return dict()

@bp.route('/register', methods=['POST'])
def register():
	""" This method is invoked by a javascript component when the user
		is trying to register the passkey. The credential is passed from
		request directly to relock service.
	"""
	if request.method == 'POST' and request.json.get('credential'):
		return request.device.webauthn(request.json)
	return request.device.webauthn()

@bp.route('/authenticate', methods=['POST'])
def authenticate():
	""" This method is invoked by a javascript component when the user
		is trying to authenticate using passkey.
	"""
	if request.method == 'POST' and 'credential' in request.json:
		return request.device.authenticate(request.json.get('credential'))
	return request.device.authenticate()

@bp.route("/open", methods=['POST'])
def open(token=None):
	""" This method is invoked by the browser side javascript automatically, 
		when the user starts a new tab in browser durring the session.
	"""
	request.device.open(request.form.get('screen', str()),
						request.form.get('origin', str()),
						request.form.get('path', str()),
						request.form.get('xsid', str()),
						request.form.get('server', str()))
	return ('', 204)

@bp.route("/close", methods=['POST'])
def close(token=None, delay=1.5):
	""" This method is invoked by the browser at he moment of website 
		beeing unloaded from the browser tab. The browser sends a hash
		specyfic to the closed tab, if this is a last tab on browser 
		side the server will close a session and logout a user.
	"""
	request.device.close(request.form.get('screen', str()),
						 request.form.get('origin', str()),
						 request.form.get('path', str()))
	return ('', 204)

@bp.route("/exchange", methods=['POST'])
def exchange():
	""" This method is invoked by javascript at the moment when the browser
		has no registered keys for the domain. Before establish of trust the 
		browser and the server need to agree the secret material of the key.
	"""
	if not isinstance(request.json.get('key'), str):
		return dict(error='The key should be an hexlified bytes')
	if not isinstance(request.json.get('hash'), str):
		return dict(error='The hash should be an hexlified bytes')
	if keys := request.device.exchange(request.json.get('key'),
									   request.json.get('hash'),
									   request.json.get('xsid'),
									   request.json.get('screen')):
		return keys
	return dict(error='Key agreement failure.')

@bp.route("/validate", methods=['POST'])
def validate(token=None, signature=None):
	""" This route is invoked by a browser every time when the transient
		key has been rotated and/or whenever the confirmation of the keys 
		is needed. The parameters are passed in headers and/or as a json 
		message directly to the device object and next by a SDK connector 
		to the relock service.
	"""
	if keys := request.device.validate(request.json.get('screen', str()),
									   request.json.get('nonce', str()),
									   request.headers.get('X-Key-Token', str()),
									   request.headers.get('X-Key-Signature', str())):
		return keys
	return dict(status=False)

@bp.route('/relock.js', methods=['GET'])
def js(minified=True):
	""" This route is invoked manually, if relock javascript is embeded 
		to web page. To reduce the processing time the compiled data from 
		the relock service are assigned to the object in the current 	
		server process memory. Local data are modyficated only when 
		server-side returns different checksum of the compiled javascript 
		file.
	"""
	# print(app.extensions.get('relock_id', str()) and app.extensions.get('relock_id', str())[:-2] in \
	#       request.headers.get('If-None-Match', str()))
	if app.extensions.get('relock_id', str()) and \
	   app.extensions.get('relock_id', str())[:-2] in \
	   request.headers.get('If-None-Match', str()):
	   #: If request headers contains If-None-Match confirming
	   #: actual ETag hash there is no need to send the entire
	   #: javascript content.
		return ('', 304)
	# print(request.headers.get('If-None-Match', str()))
	if not 'relock_id' in app.extensions or app.config.get('DEBUG',False):
		if response := request.device.js(id=app.extensions.get('relock_id', str()),
										 minified=True,
										 debug=app.config.get('DEBUG',False),
										 host=app.config.get('SERVER_HOST')):
			if response.get('js'):
				app.extensions['relock_js'] = response.get('js')
			if response.get('id'):
				app.extensions['relock_id'] = response.get('id')
				app.extensions['relock_in'] = response.get('integrity')
				app.extensions['relock_cc'] = response.get('cache', None)
	if _ := Response(app.extensions.get('relock_js', bytes()), status=200, 
													 		   content_type='text/javascript; charset=utf-8'):
		_.headers.add('Etag', app.extensions.get('relock_id', str()))
		_.headers.add('If-Match', app.extensions.get('relock_id', str()))
		if cache := app.extensions.get('relock_cc', None):
			_.headers.add('Cache-Control', 'private,max-age='+ str(cache) +',must-revalidate')
		return _

@bp.route("/clear", methods=['POST', 'GET'])
def clear():
	""" This method is invoked by a relock javascript at the moment of 
		the failure of key agreement process. If invoked it will erase
		all information about device from server repository.
	"""
	if response := request.device.clear():
		return response
	return dict(status=False)

@bp.route("/clean", methods=['POST', 'GET'])
def clean():
	""" This route is used by a server-side at the moment when key
		agreement cannot be finalized and server required complete 
		key cleanup on browser-side.
	"""
	if response := request.device.clear():
		#: response contains names to be removed from 
		#: local/session storage and also the cookies names 
		#: need to be deleted.
		if html := '<!DOCTYPE html><html lang="en"><head><meta charset="utf-8">' + \
				   '<script type="text/javascript" nonce="'+ getattr(request, '__nonce', str()) +'">' + \
				   'localStorage.removeItem("signature");' + \
				   'localStorage.removeItem("tesseract");' + \
				   'localStorage.removeItem("server");' + \
				   'localStorage.removeItem("'+ response.get('stamp', 'stamp') +'");' + \
				   'localStorage.removeItem("client");' + \
				   'localStorage.removeItem("xsid");' + \
				   'sessionStorage.removeItem("'+ response.get('cookie', 'relock') +'");' + \
				   'sessionStorage.removeItem("screen");' + \
				   'setTimeout(() => {' + \
				   'document.location=\''+ url_for('index.index') + '\';' + \
				   '}, 500);' + \
				   '</script></head><body></body></html>':
			if resp := Response(html, status=200, 
								  	  content_type='text/html; charset=utf-8'):
				resp.delete_cookie(response.get('stamp', 'stamp'), path='/')
				resp.delete_cookie(response.get('cookie', 'relock'), path='/')
				return resp
	return abort(503)